<?php
/**
 * +----------------------------------------------------------------------
 * 分解文件
 * +----------------------------------------------------------------------
 * 官网：https://www.sw-x.cn
 * +----------------------------------------------------------------------
 * 作者：小黄牛 <1731223728@qq.com>
 * +----------------------------------------------------------------------
 * 开源协议：http://www.apache.org/licenses/LICENSE-2.0
 * +----------------------------------------------------------------------
*/

class Decompose
{
    /**
     * 按行文件内容
    */
    private static $line_list = [];

    // 入口
    public static function handle($info) {
        self::speedSave($info['id'], 5);
        self::errorSave($info['id'], 0);
        self::getContext($info['file']);

        // 检测文件头注释
        self::vifHeader($info);
        // 通用文件名校验
        self::vifCurrencyFile($info);

        // 类
        if ($info['type'] == 1) {
            // 检测文件名规则
            self::vifFile($info);
            // 检测类名规则
            self::vifClass($info);
            // 检测函数名规则
            self::vifFunction($info);
        // 函数
        } else if ($info['type'] == 2) {
            // 检测函数名规则
            self::vifFunction($info);
        }

        // 配置只走下文件头就算了
        if ($info['type'] == 3) {
            // 检测文件头注释
            self::vifHeader($info);
        // 其他类型的全要走公共检测
        } else {
            // 检测变量命名规范
            self::vifParam($info);
            // 检测IF嵌套规范
            self::vifIf($info);
        }
        
        self::speedSave($info['id'], 100);
        return true;
    }

    private static function vifCurrencyFile($info) {
        $msgType = 1;
        $file = explode(DS, $info['file']);
        $name = current(explode('.', end($file)));
        $lower_name = strtolower($name);

        // 检测非法文件名
        $vif = self::deleteStr($lower_name, 3);
        if ($vif !== true) {
            self::errorSave($info['id'], $msgType, self::code($name).' 文件名包含非法字符：'.self::code($vif));
        }
        
        self::speedSave($info['id']);
    }
    private static function vifFile($info) {
        $msgType = 1;
        $file = explode(DS, $info['file']);
        $name = current(explode('.', end($file)));

        $body = self::getBody($info['file']);
        $class_pre = "/class .*?(.*?){/is"; // 类
        preg_match_all($class_pre, $body, $arr);
        $class = explode(' ', self::nbsp($arr[0][0]));
        
        if ($class[1] !== $name) {
            self::errorSave($info['id'], $msgType, '文件名与类名不一致');
        }
        self::speedSave($info['id']);
    }
    private static function vifClass($info) {
        $msgType = 2;

        $body = self::getBody($info['file']);
        $class_pre = "/class .*?(.*?){/is"; // 类
        preg_match_all($class_pre, $body, $arr);
        $arr = explode(' ', self::nbsp($arr[0][0]));
        $class = $arr[1];
        // 检测非法类名
        $vif = self::deleteStr($class);
        if ($vif !== true) {
            self::errorSave($info['id'], $msgType, self::code($class).' 类名包含非法字符：'.self::code($vif));
        }
        // 检测大驼峰格式
        $vif = ucwords($class);
        if ($class !== $vif) {
            self::errorSave($info['id'], $msgType, self::code($class).' 类名【大驼峰命名】，校验失败；应为：'.self::code($vif));
        }
        self::speedSave($info['id']);
    }
    private static function vifFunction($info) {
        $msgType = 3;

        $body = self::getBody($info['file']);
        $function_pre = "/function .*?(.*?){/is"; // 函数
        preg_match_all($function_pre, $body, $arr);
        
        if (isset($arr[0][0])) {
            $list = $arr[0];
            foreach ($list as $v) {
                $fun = str_ireplace('function ','', self::nbsp($v));
                $arr = explode('(', $fun);
                $fun = $arr[0];

                // 检测非法函数名
                $vif = self::deleteStr($fun);
                if ($vif !== true) {
                    $ret = self::getLine($v);
                    self::errorSave($info['id'], $msgType, self::code($fun.'()').' 函数名包含非法字符：'.self::code($vif), $ret['line'], $ret['code'], $ret['context']);
                }
                // 小驼峰检测
                $vif = self::smallCheck($fun);
                if ($vif !== $fun) {
                    $ret = self::getLine($v);
                    self::errorSave($info['id'], $msgType, self::code($fun.'()').' 函数名【小驼峰命名】，校验失败；应为：'.self::code($vif.'()'), $ret['line'], $ret['code'], $ret['context']);
                }
                // 函数注释
                $ret = self::getTopLine($v);
                $ret = str_replace(' ', '', $ret);
                if (
                    strpos($ret, '//') === 0 || 
                    strpos($ret, '#') === 0 || 
                    strpos($ret, '*/') === 0
                ){}else {
                    $ret = self::getLine($v);
                    self::errorSave($info['id'], $msgType, self::code($fun.'()').' 函数缺少 函数体 注释说明', $ret['line'], $ret['code'], $ret['context']);
                }
            }
        }
        
        self::speedSave($info['id']);
    }
    private static function vifHeader($info) {
        $msgType = 4;

        $status = false;
        foreach (self::$line_list as $k => $v) {
            // 只看前10行
            if ($k > 10) break;
            $v = str_replace(' ', '', $v);
            if (stripos($v, '<?') !== false) {
                continue;
            }
            if (strpos($v, '//') === 0 || strpos($v, '#') === 0 || strpos($v, '*') === 0) {
                if (isset($top)) {
                    if ($top+1 == $k) {
                        $status = true;
                    }
                    break;
                } else {
                    $top = $k; 
                    continue;  
                }
            }
        }
        if ($status == false) {
            self::errorSave($info['id'], $msgType, ' 缺少文件头部说明');
        }
        self::speedSave($info['id']);
    }
    private static function vifParam($info) {
        $msgType = 5;

        // 方法内的变量
        foreach (self::$line_list as $k => $v) {
            $param_pre = "/.*?(.*?)=/is"; // 类
            preg_match_all($param_pre, $v, $arr);
            
            if (isset($arr[0][0])) {
                $param = str_replace([' ', '='], '', $arr[0][0]);
                $ret = explode('->', $param);
                $param = $ret[0];
                
                if (strpos($param, '$') === 0) {
                    // 变量名蛇形检测
                    $vif = self::deleteStr($param, 2);
                    if ($vif !== true) {
                        $ret = self::getLine($arr[0][0]);
                        self::errorSave($info['id'], $msgType, self::code($param).' 变量名包含非法字符：'.self::code($vif), $ret['line'], $ret['code'], $ret['context']);
                        
        				self::speedSave($info['id'], 0.02);
                        continue;
                    }
                }
            }
        }

        // 函数名里的传参
        $body = self::getBody($info['file']);
        $function_pre = "/function .*?(.*?){/is"; // 函数
        preg_match_all($function_pre, $body, $arr);
        
        if (isset($arr[0][0])) {
            $list = $arr[0];
            foreach ($list as $v) {
                $fun = self::nbsp($v);
                $arr = explode('(', $v);
                $arr = explode(')', $arr[1]);
                $res = str_replace(' ', '', $arr[0]);
                if ($res) {
                    $arr = explode(',', $res);
                    foreach ($arr as $param) {
                        $res = explode('=', $param);
                        $param = $res[0];
                        // 变量名蛇形检测
                        $vif = self::deleteStr($param, 2);
                        if ($vif !== true) {
                            $ret = self::getLine($v);
                            self::errorSave($info['id'], $msgType, self::code($param).' 变量名包含非法字符：'.self::code($vif), $ret['line'], $ret['code'], $ret['context']);
                            continue;
                        }
                    }
                }
            }
        }

        self::speedSave($info['id']);
    }
    private static function vifIf($info) {
        $msgType = 6;

        // IF里的代码集合
        $body = self::getBody($info['file']);
        $if_pre = "/if.*?(.*?){.*?}/is"; // IF
        preg_match_all($if_pre, $body, $arr);
        if (isset($arr[0][0])) {
            foreach ($arr[0] as $v) {
                $num = self::recursionIf(str_replace(' ', '', $v));
                // 最大不能超过3层循环
                if ($num > 3) {
                    self::errorSave($info['id'], $msgType, self::code('if').'语句嵌套不能超过 '.self::code(3).' 层，当前已达'.self::code($num).'层', 0, '', $v);
                }
            }
        }
        self::speedSave($info['id']);
    }

    // 递归IF检测层级
    private static function recursionIf($code, $num=1) {
        $if_pre = "/if.*?(.*?){.*?}/is"; // IF
        preg_match_all($if_pre, substr($code, 2), $arr);
        if (isset($arr[0][0])) {
            $num = self::recursionIf($arr[0][0], $num+1);
        }
        return $num;
    }

    // 小驼峰语法检测
    private static function smallCheck($fun) {
        $fun = strtolower(preg_replace('/((?<=[a-z])(?=[A-Z]))/', '_', $fun));
        $arr = explode('_', $fun);
        $str = '';
        foreach ($arr as $k => $v) {
            if ($k == 0) {
                $str .= $v;
            } else {
                $str .= ucwords($v);
            }
        }
        return $str;
    }

    // 删除字符串
    private static function deleteStr($str, $type=1) {
        $pre = [' ', '$','a','b','c','d','e','f','g','h','i','j','k','l','m','n','o','p','q','r','s','t','u','v','w','x','y','z'];
        // 变量支持下划线
        if ($type == 2) {
            $pre = array_merge(['$GLOBALS', '$_GET', '$_POST', '$_SERVER', '$_REQUEST', '$_COOKIE', '$_SESSION', '_', '[', ']', '.', '"', "'", '+', '-', ':', '*', '/'], $pre);
        } else if ($type == 3) {
            $pre[] = '_';
        }
        // 变量不支持大写
        if ($type == 1) {
            $str = strtolower($str);
        }
        $str = str_replace($pre, '', $str);

        if ($str != '') {
            return $str;
        }

        return true;
    }

    // 多个空格保留一个
    private static function nbsp($str) {
        return preg_replace('!\s+!', ' ', $str);
    }

    // 加入关键字高亮标签
    private static function code($code) {
        return '<code>'.$code.'</code>';
    }

    // 更新审查原因
    private static function errorSave($id, $type, $msg='', $line=0, $code='', $context='') {
        $file = Runtime.'error'.DS.$id.'.log';
        if (!file_exists($file) || $type == 0) {
            $array = [];
        } else {
            $json = file_get_contents($file);
            $array = json_decode($json, true);
        }
        if ($type != 0) {
            $array[] = [
                'line' => $line,
                'type' => $type,
                'msg' => $msg,
                'context' => $context,
                'code' => $code,
            ];
        }
        
        file_put_contents($file, json_encode($array));
        // var_dump($array);
    }
    // 读取审查原因
    public static function errorQuery($id) {
        $file = Runtime.'error'.DS.$id.'.log';
        if (!file_exists($file)) {
            return [];
        }
        $json = file_get_contents($file);
        return json_decode($json, true);
    }

    // 更新进度
    private static function speedSave($id, $num=null) {
        if ($num === null) {
            $num = self::speedQuery($id) + 15;
        } else if ($num < 0.5) {
        	$num = self::speedQuery($id) + $num;
        }
        if ($num > 100) $num = 100;
        file_put_contents(Runtime.'speed'.DS.$id.'.log', $num);
    }

    // 读取进度
    public static function speedQuery($id) {
        $file = Runtime.'speed'.DS.$id.'.log';
        if (file_exists($file)) return file_get_contents(Runtime.'speed'.DS.$id.'.log');

        return null;
    }

    // 读取文件类型
    public static function getType($file) {
        $body = self::getBody($file);
        $type = 4;

        $class_pre = "/class .*?(.*?){/i"; // 类
        $function_pre = "/function .*?(.*?){/i"; // 函数
        $config_pre = "/return .*?(.*?)\[/i"; // 配置
        
        preg_match_all($class_pre, $body, $arr);
        if (isset($arr[0][0])) {
           return 1;
        }
        preg_match_all($function_pre, $body, $arr);
        if (isset($arr[0][0])) {
           return 2;
        }
        preg_match_all($config_pre, $body, $arr);
        if (isset($arr[0][0])) {
           return 3;
        }

        return $type;
    }

    // 读取文件内容
    private static function getBody($file) {
        return file_get_contents($file);
    }

    // 初始化行代码
    private static function getContext($file) {
        $handle = fopen($file, 'r');
        if (!$handle) {
            return $this->_json('500', '文件打开失败！');
        }
        while (false !== ($char = fgets($handle,1024))) {        //循环读取文件内容
            self::$line_list[] = $char;
        }
        fclose($handle);   
    }

    // 读取上下文
    private static function getLine($fun) {
        $line = 0;
        $code = '';
        $context = '';
        $count = count(self::$line_list);
        foreach (self::$line_list as $k => $v) {
            if (stripos($v, $fun) !== false) {
                $line = $k+1;
                $code = $v;
                if ($k != 0) {
                    $context = 'Line-'.($k).'：'.self::$line_list[$k-1];
                }
                $context .= 'Line-'.($k+1).'：'.self::$line_list[$k];
                if ($k != $count) {
                    $context .= 'Line-'.($k+2).'：'.self::$line_list[$k+1];
                }
                break;
            }
        }
        return [
            'line' => $line,
            'code' => $code,
            'context' => $context,
        ];
    }
    // 读取上一行代码
    private static function getTopLine($fun) {
        $code = '';
        foreach (self::$line_list as $k => $v) {
            if (stripos($v, $fun) !== false) {
                if ($k > 0) {
                    $code = self::$line_list[$k-1];
                }
                break;
            }
        }
        return $code;
    }
}